@@ -1,19 +1,19 @@
 
 /*
  * s3backer - FUSE-based single file backing store via Amazon S3
- * 
- * Copyright 2008-2011 Archie L. Cobbs <archie@dellroad.org>
- * 
+ *
+ * Copyright 2008-2023 Archie L. Cobbs <archie.cobbs@gmail.com>
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
@@ -37,17 +37,19 @@
 #include "s3backer.h"
 #include "block_cache.h"
 #include "ec_protect.h"
+#include "zero_cache.h"
 #include "fuse_ops.h"
 #include "http_io.h"
 #include "test_io.h"
 #include "s3b_config.h"
 #include "erase.h"
+#include "util.h"
 
 #define BLOCKS_PER_DOT          0x100
-#define MAX_QUEUE_LENGTH        1000
+#define MAX_QUEUE_LENGTH        100000
 #define NUM_ERASURE_THREADS     25
 
-/* Erasure state */
+// Erasure state
 struct erase_state {
     struct s3backer_store       *s3b;
     s3b_block_t                 queue[MAX_QUEUE_LENGTH];
@@ -56,13 +58,15 @@ struct erase_state {
     int                         quiet;
     int                         stopping;
     uintmax_t                   count;
+    bitmap_t                    *seen;
     pthread_mutex_t             mutex;
     pthread_cond_t              thread_wakeup;
     pthread_cond_t              queue_not_full;
+    pthread_cond_t              queue_empty;
 };
 
-/* Internal functions */
-static void erase_list_callback(void *arg, s3b_block_t block_num);
+// Internal functions
+static block_list_func_t erase_list_callback;
 static void *erase_thread_main(void *arg);
 
 int
@@ -71,11 +75,11 @@ s3backer_erase(struct s3b_config *config)
     struct erase_state state;
     struct erase_state *const priv = &state;
     char response[10];
+    int num_threads;
     int ok = 0;
-    int i;
     int r;
 
-    /* Double check with user */
+    // Double check with user
     if (!config->force) {
         warnx("`--erase' flag given: erasing all blocks in %s", config->description);
         fprintf(stderr, "s3backer: is this correct? [y/N] ");
@@ -90,7 +94,7 @@ s3backer_erase(struct s3b_config *config)
         }
     }
 
-    /* Initialize state */
+    // Initialize state
     memset(priv, 0, sizeof(*priv));
     priv->quiet = config->quiet;
     if ((r = pthread_mutex_init(&priv->mutex, NULL)) != 0) {
@@ -105,48 +109,61 @@ s3backer_erase(struct s3b_config *config)
         warnx("pthread_cond_init: %s", strerror(r));
         goto fail2;
     }
-    for (i = 0; i < NUM_ERASURE_THREADS; i++) {
-        if ((r = pthread_create(&priv->threads[i], NULL, erase_thread_main, priv)) != 0)
-            goto fail3;
+    if ((r = pthread_cond_init(&priv->queue_empty, NULL)) != 0) {
+        warnx("pthread_cond_init: %s", strerror(r));
+        goto fail3;
+    }
+    if ((priv->seen = bitmap_init(config->num_blocks, 0)) == NULL) {
+        r = errno;
+        warnx("calloc: %s", strerror(r));
+        goto fail4;
+    }
+    for (num_threads = 0; num_threads < NUM_ERASURE_THREADS; num_threads++) {
+        if ((r = pthread_create(&priv->threads[num_threads], NULL, erase_thread_main, priv)) != 0)
+            goto fail5;
     }
 
-    /* Logging */
+    // Logging
     if (!config->quiet) {
         fprintf(stderr, "s3backer: erasing non-zero blocks...");
         fflush(stderr);
     }
 
-    /* Create temporary lower layer */
-    if ((priv->s3b = config->test ? test_io_create(&config->http_io) : http_io_create(&config->http_io)) == NULL) {
+    // Create temporary lower layer
+    if ((priv->s3b = config->test ? test_io_create(&config->test_io) : http_io_create(&config->http_io)) == NULL) {
         warnx(config->test ? "test_io_create" : "http_io_create");
-        goto fail3;
+        goto fail5;
     }
 
-    /* Iterate over non-zero blocks */
-    if ((r = (*priv->s3b->list_blocks)(priv->s3b, erase_list_callback, priv)) != 0) {
+    // Iterate over non-zero blocks
+    if ((r = (*priv->s3b->survey_non_zero)(priv->s3b, erase_list_callback, priv)) != 0) {
         warnx("can't list blocks: %s", strerror(r));
-        goto fail3;
+        goto fail5;
     }
 
-    /* Clear mount token */
+    // Wait for queue to drain
+    pthread_mutex_lock(&priv->mutex);
+    while (priv->qlen > 0)
+        pthread_cond_wait(&priv->queue_empty, &priv->mutex);
+    CHECK_RETURN(pthread_mutex_unlock(&priv->mutex));
+
+    // Clear mount token
     if ((r = (*priv->s3b->set_mount_token)(priv->s3b, NULL, 0)) != 0) {
         warnx("can't clear mount token: %s", strerror(r));
-        goto fail3;
+        goto fail5;
     }
 
-    /* Success */
+    // Success
     ok = 1;
 
-    /* Clean up */
-fail3:
+    // Clean up
+fail5:
     pthread_mutex_lock(&priv->mutex);
     priv->stopping = 1;
     pthread_cond_broadcast(&priv->thread_wakeup);
-    pthread_mutex_unlock(&priv->mutex);
-    for (i = 0; i < NUM_ERASURE_THREADS; i++) {
-        if (priv->threads[i] == (pthread_t)0)
-            continue;
-        if ((r = pthread_join(priv->threads[i], NULL)) != 0)
+    CHECK_RETURN(pthread_mutex_unlock(&priv->mutex));
+    while (num_threads > 0) {
+        if ((r = pthread_join(priv->threads[--num_threads], NULL)) != 0)
             warnx("pthread_join: %s", strerror(r));
     }
     if (priv->s3b != NULL) {
@@ -154,8 +171,13 @@ fail3:
             fprintf(stderr, "done\n");
             warnx("erased %ju non-zero blocks", priv->count);
         }
+        (*priv->s3b->shutdown)(priv->s3b);
         (*priv->s3b->destroy)(priv->s3b);
     }
+    bitmap_free(&priv->seen);
+fail4:
+    pthread_cond_destroy(&priv->queue_empty);
+fail3:
     pthread_cond_destroy(&priv->queue_not_full);
 fail2:
     pthread_cond_destroy(&priv->thread_wakeup);
@@ -165,17 +187,25 @@ fail0:
     return ok ? 0 : -1;
 }
 
-static void
-erase_list_callback(void *arg, s3b_block_t block_num)
+static int
+erase_list_callback(void *arg, const s3b_block_t *block_nums, u_int num_blocks)
 {
     struct erase_state *const priv = arg;
 
     pthread_mutex_lock(&priv->mutex);
-    while (priv->qlen == MAX_QUEUE_LENGTH)
-        pthread_cond_wait(&priv->queue_not_full, &priv->mutex);
-    priv->queue[priv->qlen++] = block_num;
-    pthread_cond_signal(&priv->thread_wakeup);
-    pthread_mutex_unlock(&priv->mutex);
+    while (num_blocks-- > 0) {
+        const s3b_block_t block_num = *block_nums++;
+
+        if (bitmap_test(priv->seen, block_num))
+            continue;                                           // already reported to us
+        bitmap_set(priv->seen, block_num, 1);
+        while (priv->qlen == MAX_QUEUE_LENGTH)
+            pthread_cond_wait(&priv->queue_not_full, &priv->mutex);
+        priv->queue[priv->qlen++] = block_num;
+    }
+    pthread_cond_broadcast(&priv->thread_wakeup);
+    CHECK_RETURN(pthread_mutex_unlock(&priv->mutex));
+    return 0;
 }
 
 static void *
@@ -185,51 +215,53 @@ erase_thread_main(void *arg)
     s3b_block_t block_num;
     int r;
 
-    /* Acquire lock */
+    // Acquire lock
     pthread_mutex_lock(&priv->mutex);
 
-    /* Erase blocks until there are no more */
+    // Erase blocks until there are no more
     while (1) {
 
-        /* Is there a block to erase? */
+        // Is there a block to erase?
         if (priv->qlen > 0) {
 
-            /* Grab next bock */
+            // Grab next bock
             if (priv->qlen == MAX_QUEUE_LENGTH)
-                pthread_cond_signal(&priv->queue_not_full);
+                pthread_cond_broadcast(&priv->queue_not_full);
             block_num = priv->queue[--priv->qlen];
+            if (priv->qlen == 0)
+                pthread_cond_signal(&priv->queue_empty);
 
-            /* Do block deletion */
-            pthread_mutex_unlock(&priv->mutex);
+            // Do block deletion
+            CHECK_RETURN(pthread_mutex_unlock(&priv->mutex));
             r = (*priv->s3b->write_block)(priv->s3b, block_num, NULL, NULL, NULL, NULL);
             pthread_mutex_lock(&priv->mutex);
 
-            /* Check for error */
+            // Check for error
             if (r != 0) {
                 warnx("can't delete block %0*jx: %s", S3B_BLOCK_NUM_DIGITS, (uintmax_t)block_num, strerror(r));
                 continue;
             }
 
-            /* Update count and output a dot */
+            // Update count and output a dot
             if ((++priv->count % BLOCKS_PER_DOT) == 0 && !priv->quiet) {
                 fprintf(stderr, ".");
                 fflush(stderr);
             }
 
-            /* Spin again */
+            // Spin again
             continue;
         }
 
-        /* Are we done? */
+        // Are we done?
         if (priv->stopping)
             break;
 
-        /* Wait for something to do */
+        // Wait for something to do
         pthread_cond_wait(&priv->thread_wakeup, &priv->mutex);
     }
 
-    /* Done */
-    pthread_mutex_unlock(&priv->mutex);
+    // Done
+    CHECK_RETURN(pthread_mutex_unlock(&priv->mutex));
     return NULL;
 }
 
