@@ -1,19 +1,19 @@
 
 /*
  * s3backer - FUSE-based single file backing store via Amazon S3
- * 
- * Copyright 2008-2011 Archie L. Cobbs <archie@dellroad.org>
- * 
+ *
+ * Copyright 2008-2023 Archie L. Cobbs <archie.cobbs@gmail.com>
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
@@ -36,10 +36,14 @@
 
 #include "s3backer.h"
 #include "block_cache.h"
+#include "block_part.h"
 #include "ec_protect.h"
+#include "zero_cache.h"
 #include "fuse_ops.h"
 #include "http_io.h"
+#include "test_io.h"
 #include "s3b_config.h"
+#include "util.h"
 
 /****************************************************************************
  *                              DEFINITIONS                                 *
@@ -49,7 +53,7 @@
 #define FILE_INODE      2
 #define STATS_INODE     3
 
-/* Represents an open 'stats' file */
+// Represents an open 'stats' file
 struct stat_file {
     char    *buf;           // note: not necessarily nul-terminated
     size_t  len;            // length of string in 'buf'
@@ -57,22 +61,11 @@ struct stat_file {
     int     memerr;         // we got a memory error
 };
 
-/* Private information */
-struct fuse_ops_private {
-    struct s3backer_store   *s3b;
-    u_int                   block_bits;
-    off_t                   file_size;
-    time_t                  start_time;
-    time_t                  file_atime;
-    time_t                  file_mtime;
-    time_t                  stats_atime;
-};
-
 /****************************************************************************
  *                          FUNCTION DECLARATIONS                           *
  ****************************************************************************/
 
-/* FUSE functions */
+// FUSE functions
 static void *fuse_op_init(struct fuse_conn_info *conn);
 static void fuse_op_destroy(void *data);
 static int fuse_op_getattr(const char *path, struct stat *st);
@@ -94,11 +87,11 @@ static int fuse_op_unlink(const char *path);
 static int fuse_op_fallocate(const char *path, int mode, off_t offset, off_t len, struct fuse_file_info *fi);
 #endif
 
-/* Attribute functions */
+// Attribute functions
 static void fuse_op_getattr_file(struct fuse_ops_private *priv, struct stat *st);
 static void fuse_op_getattr_stats(struct fuse_ops_private *priv, struct stat_file *sfile, struct stat *st);
 
-/* Stats functions */
+// Stats functions
 static struct stat_file *fuse_op_stats_create(struct fuse_ops_private *priv);
 static void fuse_op_stats_destroy(struct stat_file *sfile);
 static printer_t fuse_op_stats_printer;
@@ -107,7 +100,7 @@ static printer_t fuse_op_stats_printer;
  *                          VARIABLE DEFINITIONS                            *
  ****************************************************************************/
 
-/* FUSE operations */
+// FUSE operations
 const struct fuse_operations s3backer_fuse_ops = {
     .init       = fuse_op_init,
     .destroy    = fuse_op_destroy,
@@ -128,7 +121,7 @@ const struct fuse_operations s3backer_fuse_ops = {
 #endif
 };
 
-/* Configuration and underlying s3backer_store */
+// Configuration and underlying s3backer_store
 static struct fuse_ops_conf *config;
 static struct fuse_ops_private *the_priv;
 
@@ -139,28 +132,42 @@ static struct fuse_ops_private *the_priv;
 const struct fuse_operations *
 fuse_ops_create(struct fuse_ops_conf *config0, struct s3backer_store *s3b)
 {
-    /* Sanity check */
+    // Sanity check
     assert(config0 != NULL);
     assert(s3b != NULL);
 
-    /* Prevent duplicate invocation */
+    // Prevent duplicate invocation
     if (config != NULL || the_priv != NULL) {
         (*config0->log)(LOG_ERR, "fuse_ops_create(): duplicate invocation");
         return NULL;
     }
+    config = config0;
 
-    /* Create private structure */
+    // Create private structure
     if ((the_priv = calloc(1, sizeof(*the_priv))) == NULL) {
         (*config->log)(LOG_ERR, "fuse_ops_create(): %s", strerror(errno));
         return NULL;
     }
     the_priv->s3b = s3b;
+    if ((the_priv->block_part = block_part_create(config->block_size, config->num_blocks)) == NULL) {
+        (*config->log)(LOG_ERR, "fuse_ops_create(): %s", strerror(errno));
+        free(the_priv);
+        return NULL;
+    }
 
-    /* Now we're ready */
-    config = config0;
+    // Now we're ready
     return &s3backer_fuse_ops;
 }
 
+void
+fuse_ops_destroy(void)
+{
+    struct fuse_ops_private *const priv = the_priv;
+
+    block_part_destroy(&priv->block_part);
+    fuse_op_destroy(priv);
+}
+
 /****************************************************************************
  *                    FUSE OPERATION FUNCTIONS                              *
  ****************************************************************************/
@@ -172,25 +179,28 @@ fuse_op_init(struct fuse_conn_info *conn)
     struct fuse_ops_private *const priv = the_priv;
     int r;
 
-    /* Sanity check */
+    // Sanity check
     assert(priv != NULL);
     assert(priv->s3b != NULL);
 
-    /* Initialize */
+    // Initialize
     priv->block_bits = ffs(config->block_size) - 1;
     priv->start_time = time(NULL);
     priv->file_atime = priv->start_time;
     priv->file_mtime = priv->start_time;
     priv->stats_atime = priv->start_time;
-    priv->file_size = config->num_blocks * config->block_size;
+    priv->file_size = (off_t)config->num_blocks * config->block_size;
 
-    /* Startup background threads now that we have fork()'d */
+    // Apply process tweaks
+    apply_process_tweaks();
+
+    // Startup background threads now that we have fork()'d
     if ((r = (*priv->s3b->create_threads)(priv->s3b)) != 0) {
         (*config->log)(LOG_ERR, "fuse_op_init(): can't create threads: %s", strerror(errno));
         return NULL;
     }
 
-    /* Done */
+    // Done
     (*config->log)(LOG_INFO, "mounting %s", s3bconf->mount);
     return priv;
 }
@@ -203,26 +213,24 @@ fuse_op_destroy(void *data)
     struct s3b_config *const s3bconf = config->s3bconf;
     int r;
 
-    /* Sanity check */
+    // Sanity check
     if (priv == NULL || s3b == NULL)
         return;
     (*config->log)(LOG_INFO, "unmount %s: initiated", s3bconf->mount);
 
-    /* Flush dirty data */
-    if (!config->read_only) {
-        (*config->log)(LOG_INFO, "unmount %s: flushing dirty data", s3bconf->mount);
-        if ((r = (*s3b->flush)(s3b)) != 0)
-            (*config->log)(LOG_ERR, "unmount %s: flushing filesystem failed: %s", s3bconf->mount, strerror(r));
-    }
+    // Shutdown (flush dirty data)
+    (*config->log)(LOG_INFO, "unmount %s: shutting down filesystem", s3bconf->mount);
+    if ((r = (*s3b->shutdown)(s3b)) != 0)
+        (*config->log)(LOG_ERR, "unmount %s: filesystem shutdown failed: %s", s3bconf->mount, strerror(r));
 
-    /* Clear mount token */
+    // Clear mount token
     if (!config->read_only) {
         (*config->log)(LOG_INFO, "unmount %s: clearing mount token", s3bconf->mount);
         if ((r = (*s3b->set_mount_token)(s3b, NULL, 0)) != 0)
             (*config->log)(LOG_ERR, "unmount %s: clearing mount token failed: %s", s3bconf->mount, strerror(r));
     }
 
-    /* Shutdown */
+    // Destroy
     (*s3b->destroy)(s3b);
     (*config->log)(LOG_INFO, "unmount %s: completed", s3bconf->mount);
     free(priv);
@@ -289,7 +297,7 @@ fuse_op_getattr_file(struct fuse_ops_private *priv, struct stat *st)
     st->st_gid = config->gid;
     st->st_size = priv->file_size;
     st->st_blksize = config->block_size;
-    st->st_blocks = config->num_blocks;
+    st->st_blocks = (off_t)config->num_blocks;
     st->st_atime = priv->file_atime;
     st->st_mtime = priv->file_mtime;
     st->st_ctime = priv->start_time;
@@ -341,11 +349,11 @@ fuse_op_open(const char *path, struct fuse_file_info *fi)
 {
     struct fuse_ops_private *const priv = (struct fuse_ops_private *)fuse_get_context()->private_data;
 
-    /* Sanity check */
+    // Sanity check
     if (priv == NULL)
         return -ENOENT;
 
-    /* Backed file */
+    // Backed file
     if (*path == '/' && strcmp(path + 1, config->filename) == 0) {
         fi->fh = 0;
         priv->file_atime = time(NULL);
@@ -354,7 +362,7 @@ fuse_op_open(const char *path, struct fuse_file_info *fi)
         return 0;
     }
 
-    /* Stats file */
+    // Stats file
     if (*path == '/' && config->print_stats != NULL && strcmp(path + 1, config->stats_filename) == 0) {
         struct stat_file *sfile;
 
@@ -366,7 +374,7 @@ fuse_op_open(const char *path, struct fuse_file_info *fi)
         return 0;
     }
 
-    /* Unknown file */
+    // Unknown file
     return -ENOENT;
 }
 
@@ -382,17 +390,14 @@ fuse_op_release(const char *path, struct fuse_file_info *fi)
 }
 
 static int
-fuse_op_read(const char *path, char *buf, size_t size, off_t offset,
-    struct fuse_file_info *fi)
+fuse_op_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
 {
     struct fuse_ops_private *const priv = (struct fuse_ops_private *)fuse_get_context()->private_data;
-    const u_int mask = config->block_size - 1;
+    struct boundary_info info;
     size_t orig_size = size;
-    s3b_block_t block_num;
-    size_t num_blocks;
     int r;
 
-    /* Handle stats file */
+    // Handle stats file
     if (fi->fh != 0) {
         struct stat_file *const sfile = (struct stat_file *)(uintptr_t)fi->fh;
 
@@ -405,7 +410,7 @@ fuse_op_read(const char *path, char *buf, size_t size, off_t offset,
         return size;
     }
 
-    /* Check for end of file */
+    // Check for end of file
     if (offset > priv->file_size) {
         (*config->log)(LOG_ERR, "read offset=0x%jx size=0x%jx out of range", (uintmax_t)offset, (uintmax_t)size);
         return -ESPIPE;
@@ -415,64 +420,40 @@ fuse_op_read(const char *path, char *buf, size_t size, off_t offset,
         orig_size = size;
     }
 
-    /* Read first block fragment (if any) */
-    if ((offset & mask) != 0) {
-        size_t fragoff = (size_t)(offset & mask);
-        size_t fraglen = (size_t)config->block_size - fragoff;
-
-        if (fraglen > size)
-            fraglen = size;
-        block_num = offset >> priv->block_bits;
-        if ((r = (*priv->s3b->read_block_part)(priv->s3b, block_num, fragoff, fraglen, buf)) != 0)
+    // Calculate what bits to read, then read them
+    calculate_boundary_info(&info, config->block_size, buf, size, offset);
+    if (info.header.length > 0 && (r = block_part_read_block_part(priv->s3b, priv->block_part, &info.header)) != 0)
+        return -r;
+    while (info.mid_block_count-- > 0) {
+        if ((r = (*priv->s3b->read_block)(priv->s3b, info.mid_block_start++, info.mid_data, NULL, NULL, 0)) != 0)
             return -r;
-        buf += fraglen;
-        offset += fraglen;
-        size -= fraglen;
+        info.mid_data += config->block_size;
     }
+    if (info.footer.length > 0 && (r = block_part_read_block_part(priv->s3b, priv->block_part, &info.footer)) != 0)
+        return -r;
 
-    /* Get block number and count */
-    block_num = offset >> priv->block_bits;
-    num_blocks = size >> priv->block_bits;
-
-    /* Read intermediate complete blocks */
-    while (num_blocks-- > 0) {
-        if ((r = (*priv->s3b->read_block)(priv->s3b, block_num++, buf, NULL, NULL, 0)) != 0)
-            return -r;
-        buf += config->block_size;
-    }
-
-    /* Read last block fragment (if any) */
-    if ((size & mask) != 0) {
-        const size_t fraglen = size & mask;
-
-        if ((r = (*priv->s3b->read_block_part)(priv->s3b, block_num, 0, fraglen, buf)) != 0)
-            return -r;
-    }
-
-    /* Done */
+    // Done
     priv->file_atime = time(NULL);
     return orig_size;
 }
 
-static int fuse_op_write(const char *path, const char *buf, size_t size,
-    off_t offset, struct fuse_file_info *fi)
+static int
+fuse_op_write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
 {
     struct fuse_ops_private *const priv = (struct fuse_ops_private *)fuse_get_context()->private_data;
-    const u_int mask = config->block_size - 1;
+    struct boundary_info info;
     size_t orig_size = size;
-    s3b_block_t block_num;
-    size_t num_blocks;
     int r;
 
-    /* Handle read-only flag */
+    // Handle read-only flag
     if (config->read_only)
         return -EROFS;
 
-    /* Handle stats file */
+    // Handle stats file
     if (fi->fh != 0)
         return -EINVAL;
 
-    /* Check for end of file */
+    // Check for end of file
     if (offset > priv->file_size) {
         (*config->log)(LOG_ERR, "write offset=0x%jx size=0x%jx out of range", (uintmax_t)offset, (uintmax_t)size);
         return -ESPIPE;
@@ -482,45 +463,23 @@ static int fuse_op_write(const char *path, const char *buf, size_t size,
         orig_size = size;
     }
 
-    /* Handle request to write nothing */
+    // Handle request to write nothing
     if (size == 0)
         return 0;
 
-    /* Write first block fragment (if any) */
-    if ((offset & mask) != 0) {
-        size_t fragoff = (size_t)(offset & mask);
-        size_t fraglen = (size_t)config->block_size - fragoff;
-
-        if (fraglen > size)
-            fraglen = size;
-        block_num = offset >> priv->block_bits;
-        if ((r = (*priv->s3b->write_block_part)(priv->s3b, block_num, fragoff, fraglen, buf)) != 0)
+    // Calculate what bits to write, then write them
+    calculate_boundary_info(&info, config->block_size, buf, size, offset);
+    if (info.header.length > 0 && (r = block_part_write_block_part(priv->s3b, priv->block_part, &info.header)) != 0)
+        return -r;
+    while (info.mid_block_count-- > 0) {
+        if ((r = (*priv->s3b->write_block)(priv->s3b, info.mid_block_start++, info.mid_data, NULL, NULL, NULL)) != 0)
             return -r;
-        buf += fraglen;
-        offset += fraglen;
-        size -= fraglen;
+        info.mid_data += config->block_size;
     }
+    if (info.footer.length > 0 && (r = block_part_write_block_part(priv->s3b, priv->block_part, &info.footer)) != 0)
+        return -r;
 
-    /* Get block number and count */
-    block_num = offset >> priv->block_bits;
-    num_blocks = size >> priv->block_bits;
-
-    /* Write intermediate complete blocks */
-    while (num_blocks-- > 0) {
-        if ((r = (*priv->s3b->write_block)(priv->s3b, block_num++, buf, NULL, NULL, NULL)) != 0)
-            return -r;
-        buf += config->block_size;
-    }
-
-    /* Write last block fragment (if any) */
-    if ((size & mask) != 0) {
-        const size_t fraglen = size & mask;
-
-        if ((r = (*priv->s3b->write_block_part)(priv->s3b, block_num, 0, fraglen, buf)) != 0)
-            return -r;
-    }
-
-    /* Done */
+    // Done
     priv->file_mtime = time(NULL);
     return orig_size;
 }
@@ -530,7 +489,7 @@ fuse_op_statfs(const char *path, struct statvfs *st)
 {
     st->f_bsize = config->block_size;
     st->f_frsize = config->block_size;
-    st->f_blocks = config->num_blocks;
+    st->f_blocks = (off_t)config->num_blocks;
     st->f_bfree = 0;
     st->f_bavail = 0;
     st->f_files = 3;
@@ -548,6 +507,18 @@ fuse_op_truncate(const char *path, off_t size)
 static int
 fuse_op_flush(const char *path, struct fuse_file_info *fi)
 {
+    struct fuse_ops_private *const priv = (struct fuse_ops_private *)fuse_get_context()->private_data;
+    int r;
+
+    // Ignore if stats file
+    if (fi->fh != 0)
+        return 0;
+
+    // Flush ALL dirty blocks
+    if ((r = (*priv->s3b->flush_blocks)(priv->s3b, NULL, 0, 0)) != 0)
+        return -r;
+
+    // Done
     return 0;
 }
 
@@ -560,7 +531,7 @@ fuse_op_fsync(const char *path, int isdatasync, struct fuse_file_info *fi)
 static int
 fuse_op_unlink(const char *path)
 {
-    /* Handle stats file */
+    // Handle stats file
     if (*path == '/' && strcmp(path + 1, config->stats_filename) == 0) {
         if (config->clear_stats == NULL)
             return -EOPNOTSUPP;
@@ -568,7 +539,7 @@ fuse_op_unlink(const char *path)
         return 0;
     }
 
-    /* Not supported */
+    // Not supported
     return -EOPNOTSUPP;
 }
 
@@ -578,24 +549,21 @@ static int
 fuse_op_fallocate(const char *path, int mode, off_t offset, off_t len, struct fuse_file_info *fi)
 {
     struct fuse_ops_private *const priv = (struct fuse_ops_private *)fuse_get_context()->private_data;
-    const u_int mask = config->block_size - 1;
+    struct boundary_info info;
     size_t size = (size_t)len;
-    s3b_block_t block_num;
-    void *zero_block;
-    size_t num_blocks;
     int r;
 
-    /* Handle stats file */
+    // Handle stats file
     if (fi->fh != 0)
         return -EOPNOTSUPP;
 
-    /* Sanity check */
+    // Sanity check
     if (offset < 0 || len <= 0)
         return -EINVAL;
     if (offset + len > priv->file_size)
         return -ENOSPC;
 
-    /* Handle request */
+    // Handle request
     if ((mode & FALLOC_FL_PUNCH_HOLE) == 0)
         return 0;
 /*
@@ -603,51 +571,19 @@ fuse_op_fallocate(const char *path, int mode, off_t offset, off_t len, struct fu
         return -EINVAL;
 */
 
-    /* Create an empty block */
-    if ((zero_block = calloc(1, config->block_size)) == NULL)
-        return -ENOMEM;
-
-    /* Write first block fragment (if any) */
-    if ((offset & mask) != 0) {
-        size_t fragoff = (size_t)(offset & mask);
-        size_t fraglen = (size_t)config->block_size - fragoff;
-
-        if (fraglen > size)
-            fraglen = size;
-        block_num = offset >> priv->block_bits;
-        if ((r = (*priv->s3b->write_block_part)(priv->s3b, block_num, fragoff, fraglen, zero_block)) != 0) {
-            free(zero_block);
+    // Calculate what bits to write, then write them
+    calculate_boundary_info(&info, config->block_size, NULL, size, offset);
+    if (info.header.length > 0 && (r = block_part_write_block_part(priv->s3b, priv->block_part, &info.header)) != 0)
+        return -r;
+    while (info.mid_block_count-- > 0) {
+        if ((r = (*priv->s3b->write_block)(priv->s3b, info.mid_block_start++, NULL, NULL, NULL, NULL)) != 0)
             return -r;
-        }
-        offset += fraglen;
-        size -= fraglen;
     }
+    if (info.footer.length > 0 && (r = block_part_write_block_part(priv->s3b, priv->block_part, &info.footer)) != 0)
+        return -r;
 
-    /* Get block number and count */
-    block_num = offset >> priv->block_bits;
-    num_blocks = size >> priv->block_bits;
-
-    /* Write intermediate complete blocks */
-    while (num_blocks-- > 0) {
-        if ((r = (*priv->s3b->write_block)(priv->s3b, block_num++, NULL, NULL, NULL, NULL)) != 0) {
-            free(zero_block);
-            return -r;
-        }
-    }
-
-    /* Write last block fragment (if any) */
-    if ((size & mask) != 0) {
-        const size_t fraglen = size & mask;
-
-        if ((r = (*priv->s3b->write_block_part)(priv->s3b, block_num, 0, fraglen, zero_block)) != 0) {
-            free(zero_block);
-            return -r;
-        }
-    }
-
-    /* Done */
+    // Done
     priv->file_mtime = time(NULL);
-    free(zero_block);
     return 0;
 }
 #endif
@@ -688,12 +624,12 @@ fuse_op_stats_printer(void *prarg, const char *fmt, ...)
     size_t remain;
     int added;
 
-    /* Bail if no memory */
+    // Bail if no memory
     if (sfile->memerr)
         return;
 
 again:
-    /* Append to string buffer */
+    // Append to string buffer
     remain = sfile->bufsiz - sfile->len;
     va_start(args, fmt);
     added = vsnprintf(sfile->buf + sfile->len, sfile->bufsiz - sfile->len, fmt, args);
@@ -703,7 +639,7 @@ again:
         return;
     }
 
-    /* We need a bigger buffer */
+    // We need a bigger buffer
     new_bufsiz = ((sfile->bufsiz + added + 1023) / 1024) * 1024;
     if ((new_buf = realloc(sfile->buf, new_bufsiz)) == NULL) {
         sfile->memerr = 1;
