@@ -1,19 +1,19 @@
 
 /*
  * s3backer - FUSE-based single file backing store via Amazon S3
- * 
- * Copyright 2008-2011 Archie L. Cobbs <archie@dellroad.org>
- * 
+ *
+ * Copyright 2008-2023 Archie L. Cobbs <archie.cobbs@gmail.com>
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
@@ -34,61 +34,72 @@
  * also delete it here.
  */
 
-/* Upload/download indexes */
+// Upload/download indexes
 #define HTTP_DOWNLOAD       0
 #define HTTP_UPLOAD         1
 
-/* Authentication types */
+// Authentication types
 #define AUTH_VERSION_AWS2   "aws2"
 #define AUTH_VERSION_AWS4   "aws4"
 
-/* Storage classes */
+// Storage classes
 #define STORAGE_CLASS_STANDARD              "STANDARD"
 #define STORAGE_CLASS_STANDARD_IA           "STANDARD_IA"
+#define STORAGE_CLASS_ONEZONE_IA            "ONEZONE_IA"
 #define STORAGE_CLASS_REDUCED_REDUNDANCY    "REDUCED_REDUNDANCY"
+#define STORAGE_CLASS_INTELLIGENT_TIERING   "INTELLIGENT_TIERING"
+#define STORAGE_CLASS_GLACIER               "GLACIER"
+#define STORAGE_CLASS_DEEP_ARCHIVE          "DEEP_ARCHIVE"
+#define STORAGE_CLASS_OUTPOSTS              "OUTPOSTS"
 
-/* Server side encryption types */
-#define REQUIRED_SSE_VALUE                  "AES256"
+// Server side encryption types
+#define SSE_AES256                          "AES256"
+#define SSE_AWS_KMS                         "aws:kms"
 
-/* Configuration info structure for http_io store */
+// Configuration info structure for http_io store
 struct http_io_conf {
-    char                *accessId;
-    char                *accessKey;
-    char                *iam_token;
-    const char          *accessType;
-    const char          *ec2iam_role;
-    const char          *storage_class;
-    const char          *authVersion;
-    const char          *baseURL;
-    const char          *region;
-    const char          *bucket;
-    const char          *prefix;
-    const char          *user_agent;
-    const char          *cacert;
-    const char          *password;
-    const char          *encryption;
-    const char          *default_ce;
-    u_int               key_length;
-    int                 debug;
-    int                 debug_http;
-    int                 quiet;
-    int                 rrs;                        // reduced redundancy storage (backward compat.)
-    int                 compress;                   // zlib compression level
-    int                 vhost;                      // use virtual host style URL
-    u_int               *nonzero_bitmap;            // is set to NULL by http_io_create()
-    int                 blockHashPrefix;
-    int                 insecure;
-    u_int               block_size;
-    off_t               num_blocks;
-    u_int               timeout;
-    u_int               initial_retry_pause;
-    u_int               max_retry_pause;
-    uintmax_t           max_speed[2];
-    log_func_t          *log;
-    const char          *sse;
+    char                    *accessId;
+    char                    *accessKey;
+    char                    *iam_token;
+    const char              *accessType;
+    const char              *ec2iam_role;
+    const char              *storage_class;
+    const char              *authVersion;
+    const char              *baseURL;
+    const char              *vhostURL;                  // "baseURL" in --vhost format
+    const char              *region;
+    const char              *bucket;
+    const char              *prefix;
+    const char              *user_agent;
+    const char              *cacert;
+    const char              *password;
+    const char              *encryption;
+    const char              *default_ce;
+    u_int                   key_length;
+    int                     debug;
+    int                     debug_http;
+    int                     http_11;                    // restrict to HTTP 1.1
+    int                     quiet;
+    int                     no_curl_cache;              // don't cache cURL handles
+    const struct comp_alg   *compress_alg;              // compression algorithm, or NULL for none
+    void                    *compress_level;            // compression level info
+    int                     vhost;                      // use virtual host style URL
+    bitmap_t                *nonzero_bitmap;            // is set to NULL by http_io_create()
+    int                     blockHashPrefix;
+    int                     insecure;
+    u_int                   block_size;
+    s3b_block_t             num_blocks;
+    int                     list_blocks_threads;
+    u_int                   timeout;
+    u_int                   initial_retry_pause;
+    u_int                   max_retry_pause;
+    uintmax_t               max_speed[2];
+    log_func_t              *log;
+    const char              *sse;
+    const char              *sse_key_id;
 };
 
-/* Statistics structure for http_io store */
+// Statistics structure for http_io store
 struct http_io_evst {
     u_int               count;                      // number of occurrences
     double              time;                       // total time taken
@@ -96,7 +107,7 @@ struct http_io_evst {
 
 struct http_io_stats {
 
-    /* Block stats */
+    // Block stats
     u_int               normal_blocks_read;
     u_int               normal_blocks_written;
     u_int               zero_blocks_read;
@@ -104,7 +115,7 @@ struct http_io_stats {
     u_int               empty_blocks_read;          // only when nonzero_bitmap != NULL
     u_int               empty_blocks_written;       // only when nonzero_bitmap != NULL
 
-    /* HTTP transfer stats */
+    // HTTP transfer stats
     struct http_io_evst http_heads;                 // total successful
     struct http_io_evst http_gets;                  // total successful
     struct http_io_evst http_puts;                  // total successful
@@ -112,14 +123,16 @@ struct http_io_stats {
     u_int               http_unauthorized;
     u_int               http_forbidden;
     u_int               http_stale;
+    u_int               http_redirect;
     u_int               http_verified;
     u_int               http_mismatch;
     u_int               http_5xx_error;
     u_int               http_4xx_error;
+    u_int               http_3xx_error;
     u_int               http_other_error;
     u_int               http_canceled_writes;
 
-    /* CURL stats */
+    // CURL stats
     u_int               curl_handles_created;
     u_int               curl_handles_reused;
     u_int               curl_timeouts;
@@ -128,18 +141,20 @@ struct http_io_stats {
     u_int               curl_out_of_memory;
     u_int               curl_other_error;
 
-    /* Retry stats */
+    // Retry stats
     u_int               num_retries;
     uint64_t            retry_delay;
 
-    /* Misc */
+    // Misc
     u_int               out_of_memory_errors;
 };
 
-/* http_io.c */
+// http_io.c
 extern struct s3backer_store *http_io_create(struct http_io_conf *config);
 extern void http_io_get_stats(struct s3backer_store *s3b, struct http_io_stats *stats);
 extern void http_io_clear_stats(struct s3backer_store *s3b);
-extern int http_io_parse_block(struct http_io_conf *config, const char *name, s3b_block_t *block_num);
-extern void http_io_format_block_hash(const struct http_io_conf *config, char *block_hash_buf, size_t bufsiz, s3b_block_t block_num);
+extern int http_io_parse_block(const char *prefix, s3b_block_t num_blocks,
+    int blockHashPrefix, const char *name, s3b_block_t *hash_valuep, s3b_block_t *block_nump);
+extern void http_io_format_block_hash(int blockHashPrefix, char *block_hash_buf, size_t bufsiz, s3b_block_t block_num);
+extern int http_io_list_blocks(struct s3backer_store *s3b, block_list_func_t *callback, void *arg);
 
