@@ -1,19 +1,19 @@
 
 /*
  * s3backer - FUSE-based single file backing store via Amazon S3
- * 
- * Copyright 2008-2011 Archie L. Cobbs <archie@dellroad.org>
- * 
+ *
+ * Copyright 2008-2023 Archie L. Cobbs <archie.cobbs@gmail.com>
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
@@ -37,6 +37,7 @@
 #include "s3backer.h"
 #include "block_cache.h"
 #include "ec_protect.h"
+#include "zero_cache.h"
 #include "fuse_ops.h"
 #include "http_io.h"
 #include "test_io.h"
@@ -53,23 +54,23 @@ s3backer_reset(struct s3b_config *config)
     int ok = 0;
     int r;
 
-    /* Logging */
+    // Logging
     if (!config->quiet)
         warnx("resetting mount token for %s", config->description);
 
-    /* Create temporary lower layer */
-    if ((s3b = config->test ? test_io_create(&config->http_io) : http_io_create(&config->http_io)) == NULL) {
+    // Create temporary lower layer
+    if ((s3b = config->test ? test_io_create(&config->test_io) : http_io_create(&config->http_io)) == NULL) {
         warnx(config->test ? "test_io_create" : "http_io_create");
         goto fail;
     }
 
-    /* Clear mount token */
+    // Clear mount token
     if ((r = (*s3b->set_mount_token)(s3b, NULL, 0)) != 0) {
         warnx("error clearing s3 mount token: %s", strerror(r));
         goto fail;
     }
 
-    /* Open disk cache file, if any, and clear the mount token there too */
+    // Open disk cache file, if any, and clear the mount token there too
     if (config->block_cache.cache_file != NULL) {
         if (stat(config->block_cache.cache_file, &cache_file_stat) == -1) {
             if (errno != ENOENT) {
@@ -77,25 +78,26 @@ s3backer_reset(struct s3b_config *config)
                 goto fail;
             }
         } else {
-            if ((r = s3b_dcache_open(&dcache, config->log, config->block_cache.cache_file,
-              config->block_cache.block_size, config->block_cache.cache_size, NULL, NULL, 0)) != 0)
+            if ((r = s3b_dcache_open(&dcache, &config->block_cache, NULL, NULL, 0)) != 0)
                 warnx("error opening cache file `%s': %s", config->block_cache.cache_file, strerror(r));
             if ((r = s3b_dcache_set_mount_token(dcache, NULL, 0)) != 0)
                 warnx("error reading mount token from `%s': %s", config->block_cache.cache_file, strerror(r));
         }
     }
 
-    /* Success */
+    // Success
     if (!config->quiet)
         warnx("done");
     ok = 1;
 
 fail:
-    /* Clean up */
+    // Clean up
     if (dcache != NULL)
         s3b_dcache_close(dcache);
-    if (s3b != NULL)
+    if (s3b != NULL) {
+        (*s3b->shutdown)(s3b);
         (*s3b->destroy)(s3b);
+    }
     return ok ? 0 : -1;
 }
 
